// @ myth - common - buffer.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk
//
// A generic file buffer class.
//
// The class BUFFER provide the user with buffered in/out. To create a
// BUFFER from an file descriptor (see open(2)) just declare it:
//    BUFFER logfile(BUFFER_READ, fd); // or
//    BUFFER line(BUFFER_RDWR);
//    line.open(fd);
//
// The BUFFER may be declared for READ, WRITE or RDWR. It must reflect
// the way the file descriptor is defined.
//
// You may also create the BUFFER from a filename:
//    BUFFER logfile(BUFFER_READ, path); // or
//    BUFFER line(BUFFER_RDWR);
//    line.open(path);
//
// You may now write and/or read to the buffer, with:
//    logfile.put("Hello", 6);
//    line.get(&i, sizeof i);
//    line.put(&c, 1); 
// 
// And output buffer will be flushed (send) when it gets full, or when
// the BUFFER is deleted. It can also be flushed by:
//    line.flush();
//
// Any errors in writing or reading will set a flag in the BUFFER. End
// of file for the input also sets a flag. These flags can be tested
// with:
//    if((line.eof() || line.GoodInput()) && line.GoodOutput) // No Error!
//

// @@ include

#include "common/buffer.h"
#include "common/error.h"
#include "common/format.h"

// @@ const

// Size used for buffered I/O.
const int BufferSize = 1024;
// Mode to create new files with.
static const int FileCreationMode = 0666;

// @@ BUFFER

// @@@ void BUFFER::create(buffer_mode mode)

void BUFFER::create(buffer_mode mode)
{
    // Common state
    BUFFER::fd = ERROR;
    BUFFER::mode = mode;
    BUFFER::IsReady = FALSE;
        
    // Input state
    current = max = 0;
    input = BUFFER_SUCCES;
    in = (mode & BUFFER_READ) ? new char[BufferSize] : NULL;
    
    // Output state
    pos = 0;
    output = SUCCES;
    out = (mode & BUFFER_WRITE) ? new char[BufferSize] : NULL;
}

// @@@ void BUFFER::put(const void* buffer, int size)
//
// Write some data to the BUFFER

void BUFFER::put(const void* buffer, int size)
{
    IsTouched = TRUE;

    if(!(mode & BUFFER_WRITE))
    {
	bug.error("BUFFER::put", format("Bad mode %d", mode));
	output = ERROR;
    }
    if(!GoodOutput())
	return;

    const char* buf = (char*) buffer;

    while(size > 0) 
    {
	out[pos] = *buf;
	pos++;
	size--;
	buf++;
	if(pos == BufferSize)
	{
	    flush();

	    if(!GoodOutput())
	    {
		bug.warning("BUFFER", "GoodOutput() is false");
		return;
	    }
	}
    }
}

// @@@ void BUFFER::get(void* buffer, int size);
//
// Read some data from the buffer

void BUFFER::get(void* buffer, int size)
{
    if(!(mode & BUFFER_READ))
    {
    	bug.error("BUFFER::put", format("Bad mode %d", mode));
	input = BUFFER_ERROR;
    }
    if(!GoodInput())
	return;

    char* buf = (char*) buffer;

    while(size)
    {
	if(max == current)
	{
	    switch(max = read(fd, in, BufferSize))
	    {
	    case ERROR:
		bug.warning("BUFFER::get", "read() failed");
		input = BUFFER_ERROR;
		return;
	    case 0:
		input = BUFFER_EOF;
		return;
	    default:
		current = 0;
	    }
	}
	*buf = in[current];
	current++;
	buf++;
	size--;
    }
    BUFFER::IsReady = (max != current);
}

// @@@ void BUFFER::open(STRING name)
//
// Connect a buffer to a file.

void BUFFER::open(STRING name)
{
    int NewMode;

    switch(mode)
    {
    case BUFFER_NONE:
	bug.error("BUFFER::open", format("open BUFFER_NONE"));
	return;
    case BUFFER_READ:
	NewMode = O_RDONLY;
	break;
    case BUFFER_WRITE:
	NewMode = O_WRONLY | O_CREAT | O_APPEND;
	break;
    case BUFFER_RDWR:
	NewMode = O_RDWR;
	break;
    case BUFFER_EXCLUSIVE:
	NewMode = O_WRONLY | O_CREAT | O_EXCL;
	break;
    default:
	bug.error("BUFFER::open", format("Unknow mode %d", mode));
	return;
    }
    BUFFER::open(::open(name(), NewMode, FileCreationMode));
}

// @@@ void BUFFER::open(int fd)
//
// Assign a file descriptor to buffer

void BUFFER::open(int fdno)
{
    if(fd == ERROR)
	fd = fdno;
    else
    {
	bug.error("BUFFER::open",
		  format("fd = %d, fdno = %d", fd, fdno));

	if(fd != fdno)
	    (void) close(fdno);
	else
	    fd = fdno;
    }
}

// @@@ void BUFFER::flush()
//
// Send all data from output buffer to file descriptor

void BUFFER::flush()
{
    if(!GoodOutput())
	return;
    
    char* buffer = out;
    
    while(pos)
    {
	int done = write(fd, buffer, pos);
	
	if(done == ERROR)
	{
	    bug.warning("BUFFER::flush", "write() failed");
	    return;
	}
	pos -= done;
	buffer += done;
    }
}

// @@@ BUFFER::BUFFER(buffer_mode mode, int fd = ERROR) 
//
// Create a buffer from a file descriptor.

BUFFER::BUFFER(buffer_mode mode, int fd = ERROR)
{
    create(mode);
    BUFFER::fd = fd;
}

// @@@ BUFFER::BUFFER(buffer_mode mode, STRING name) 
//
// Create a buffer from a filename.


BUFFER::BUFFER(buffer_mode mode, STRING name)
{
    create(mode);
    BUFFER::open(name);
}

// @@@ BUFFER::~BUFFER()
//
// Delete the buffers and close the file.


BUFFER::~BUFFER()
{
    if(mode & BUFFER_WRITE)
    {
	flush();
	delete[BufferSize] out;
    }
    if(mode & BUFFER_READ)
	delete[BufferSize] in;

    if(fd == ERROR)
	return;

    (void) close(fd);
}

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
