// @ myth - server - level.h
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ begin

#ifndef LEVEL_HEADER
#define LEVEL_HEADER

// @@ include

#include "common/config.h"
#include "common/list.h"
#include "common/map.h"
#include "server/bunch.h"
#include "server/party.h"
#include "server/global.h"

struct BINARY;
struct WAIT;
struct SLEEP;
struct RAY;
struct HI_SCORE;

// @@ const

// How many seconds between each players is allowed to hit each other
static const int TickTime = 2;

// Timeout constants.
static const int MinLevelLife = 10;
static const int MaxLevelLife = 500;

// Size of the two dimensional array of buckets.
static const int BucketSize = 25;

// Information stored in the map.
enum { MAP_MASK_BIT = 64, MAP_MASK_LIT = 8,
       MAP_MASK_ROCK = 1 + 2 + 4, MAP_OFFSET_ROCK = 0 };

// @@ FREQUENCE

// @@@ class frequence_link
//
// Private! Do not touch...

struct frequence_link
{
    object_type object;
    int chance;
    void save(BINARY* binary);
    frequence_link(BINARY* binary);
    frequence_link(object_type type, int when);
};

// @@@ class FREQUENCE

class FREQUENCE : LIST
{
    // * Chances
    int chance_monster;		// Chance for creating a monster
    int chance_object;		// Chance for creating an object
    int chance_create;		// Chance a new monster appears
    int all_monsters;		// Sum of chance for all monsters
    // * Location
    STRING where;		// Where new monsters can appear
    // * List operations
    void reset()
    { LIST::reset(); }
    void next()
    { LIST::next(); }
    void erase();
    void clear();
    bool check()
    { return LIST::current() != NULL; }
    void add(object_type object, int chance);
    frequence_link* current()
    { return (frequence_link*) LIST::current(); }
    int chance()
    { return current()->chance; }
    object_type type()
    { return current()->object; }
public:
    MONSTER* CreateMonster(int value);
    bool MonsterChance();
    bool ObjectChance();
    bool NewMonster();
    char locate();
    void save(BINARY* binary);
    void load(BINARY* binary);
    void load(PARSE* parse);
    FREQUENCE();
    ~FREQUENCE()
    { clear(); }
};

// @@ BUCKET

class BUCKET
{
    LIST BunchList;
    LIST PositionList;
public:
    BUNCH* what(coor pos);	// Return entrie at pos, or NULL
    BUNCH* create(coor pos);	// Create and return entrie at pos
    void update();		// Remove empty entries
    void save(BINARY* binary);	// Save all entries to file
    void free();		// Remove and deallocate all entries
};

// @@ EVENT

// @@@ event_link

struct event_link
{
    coor pos;
    long when;
    event_link(coor p, long w)
    { pos = p; when = w; }
};

// @@@ class EVENT

class EVENT
{
    LIST list;
    void reset()
    { list.reset(); }
    void next()
    { list.next(); }
    event_link* current()
    { return (event_link*) list.current(); }
    void insert(coor pos, long when)
    { list.insert(new event_link(pos, when)); }
    void erase();
    void clear();
public:
    void add(coor pos, long when);
    void update(LEVEL* level);
    void save(BINARY* binary);
    void load(BINARY* binary);
    EVENT()
    { }
    ~EVENT()
    { clear(); }
};

// @@ enum tick_type
//
// What happend when I MoveAllPlayers?
enum tick_type
{ TICK_WAIT, TICK_TOUCH, TICK_MOVE, TICK_FIGHT, TICK_ERROR = ERROR };

// @@ class LEVEL

class LEVEL
{
    // Action
    //
private:
    tick_type MoveAllPlayers(bool* AutoRun);
    void MoveAllMonsters();
    void CleanUpPlayerQueue(WAIT* wait, SLEEP* sleep, HI_SCORE* HiScore);
    void CleanUpMonsterQueue(MONSTER_LIST* export);
public:
    bool play(MONSTER_LIST* export, WAIT* wait,
	      SLEEP* sleep, HI_SCORE* HiScore, force flag);
    bool timeout()		// Level has timed out
    { return CountDown != ERROR && --CountDown < 1; }
    void fight(MONSTER* attack, MONSTER* defence);
    void throw(MONSTER* monster, OBJECT* object, coor target);
    void zap(MONSTER* monster, RAY* ray, coor target);
    void dig(MONSTER* monster, coor target);
    bool hit(coor pos, int power,
	     OBJECT* object, MONSTER* monster, coor last /* = NoWhere */);
    //
    // Messages
    //
    // Tell players seeing pos
    void WatchMonster(MONSTER* monster, const char* string,
		      OBJECT* object = NULL);
    void WatchPosition(coor pos, const char* string);
    void WatchPosition(coor pos, STRING string);
    void WatchEffect(coor pos, char c);
    void hiss(MONSTER* hissing);
    void talk(MONSTER* talker, const char* what);
    //
    // Population.
    //
private:
    MONSTER_LIST zoo; 		// Monsters on the level
    PLAYER_LIST party;		// Players on the level
    PLAYER_LIST sleepers;	// Sleeping players
    EVENT event;		// Events planned by the objects
    FREQUENCE frequence;	// Frequence of monsters on this level
public:
    void AddMonster(MONSTER* monster);
    void AddPlayer(PLAYER* player);
    err restore(id_type id, PLAYER* player);
    void GetMonster(MONSTER_LIST* list, coor pos);
    PLAYER* FindPlayer(MONSTER* monster);
    MONSTER* FindEnemy(MONSTER* monster);
    //
    // Objects
    //
private:
    BUCKET* bucket[BucketSize]; // Objects on the level
public:
    BUNCH* bunch(coor pos); // Return BUNCH at pos, or NULL
    BUNCH* create(coor pos); // Return BUNCH at pos, or create it
    void update(coor pos); // Remove BUNCH at pos if empty
    void AddEvent(coor pos, long when); // Something happens at pos later
    //
    // Map
    //
private:
    coor location[MaxChar];	// Identification of special locations
    MAP map;			// A map of the level
    char RockSymbol[ROCK_END];	// Symbols used for each type of rock
public:
    // Light
    bool lit(coor pos);
    void light(coor pos);
    void dark(coor pos);
    // Rock
    rock_type rock(coor pos);
    void dig(coor pos, rock_type rock);
    // Return symbol to be shown on screen.
    char symbol(coor pos, bool SeeInvisible);
    bool IsLit(coor pos);	// Return TRUE, iff there is some light source
    coor locate(int l);		// Find the location marked l.
    //
    // Administration
    //
private:
    long LastVisit;		// Last time we updated the FightPoints
    bool WatchClock();		// Check if it is time for a new update
    int CountDown;		// Time before level should be removed
    level_type MyDepth;		// What cave number?
    int MyExperience;		// How difficult is this cave?
    int SaveChance;		// How many promille chance for saving binary?
    void save(BINARY* binary); // Save level to binary file
    void load(BINARY* binary); // Load level from binary file
    char ParseValue(PARSE* parse); // Value of map location
    void LoadText(PARSE* parse); // Load level from text file
    void CreateLevel(level_type StairUp, level_type StairDown);
    void CreateCave(coor where, rock_type value, int width, int chance = 1);
    void CreatePath(coor from, coor to);
public:
    void CheckOut(MONSTER_LIST* export,	// Get rid of this level
		  WAIT* wait, SLEEP* sleep, HI_SCORE* HiScore);
    level_type depth()
    { return MyDepth; }
    int experience()
    { return MyExperience; }
    LEVEL(level_type type);
    ~LEVEL();			// Free buckets
};

// @@ end

#endif LEVEL_HEADER

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
