// @ myth - server - thing.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ include

#include "server/thing.h"
#include "common/error.h"
#include "common/binary.h"
#include "common/parse.h"
#include "server/monster.h"
#include "server/player.h"
#include "server/rnd.h"
#include "server/ray.h"
#include "common/format.h"

// @@ THING

// @@@ void THING::create(long count = 1L)

void THING::create(long count = 1L)
{
    MyLetter = 'a';
    MyValue = 1;
    holy = NORMAL;
    invisible = FALSE;
    MyCount = count;
}

// @@@ comp THING::compare(OBJECT* object)

comp THING::compare(OBJECT* object)
{
    if(!object->IsThing())
	return OBJECT::compare(object);

    THING* thing = (THING*) object;
    
    if(type() < object->type())
	return LESS_THAN;
    else if(type() > object->type())
	return GREATER_THAN;
    else if(invisible && !thing->invisible)
	return LESS_THAN;
    else if(thing->invisible && !invisible)
	return GREATER_THAN;
    else if(IsCursed() && !thing->IsCursed())
	return LESS_THAN;
    else if(thing->IsCursed() && !IsCursed())
	return GREATER_THAN;
    else if(thing->IsBlessed() && !IsBlessed())
	return LESS_THAN;
    else if(IsBlessed() && !thing->IsBlessed())
	return GREATER_THAN;
    else if(MyValue < thing->MyValue)
	return LESS_THAN;
    else if(MyValue > thing->MyValue)
	return GREATER_THAN;

    return EQUAL;
}

// @@@ STRING THING::print(print_flag flag, long count, PLAYER* player)

STRING THING::print(print_flag flag, long count, PLAYER* player)
{
    char string[MessageSize];
    string[0] = 0;
    
    if(flag & PRINT_VALUE)
    {
	const char* curse = PrintHoly();

	if(curse != NULL)
	{
	    (void) strcat(string, curse);
	    (void) strcat(string, " ");
	}
	(void) strcat(string, PrintValue());
	(void) strcat(string, " ");
    }
    const char* active = PrintActive(flag);

    if(active != NULL)
    {
	(void) strcat(string, active);
	(void) strcat(string, " ");
    }
    STRING bugfix = OBJECT::print(flag, count, player);
    const char* text = bugfix.string();

    return format("%s%s", string, text);
}

// @@@ const char* THING::PrintValue()

const char* THING::PrintValue()
{
    switch(MyValue)
    {
    case 0:
        return "useless";
    case 1: case 2: case 3: case 4:
        return "almost useless";
    case 5: case 6: case 7: case 8: case 9: case 10: case 11:
        return "average";
    case 12: case 13: case 14: case 15: case 16: case 17:
        return "good";
    case 18: case 19: case 20: case 21: case 22:
        return "very good";
    case 23: case 24:
        return "nearly perfect";
    case 25:
        return "perfect";
    default:
        return "impossible";
    }
}

// @@@ const char* THING::PrintHoly()

const char* THING::PrintHoly()
{
    if(IsBlessed() && IsCursed())
	return "strange";
    else if(IsBlessed())
	return "blessed";
    else if(IsCursed())
	return "cursed";
    else
	return NULL;
}

// @@@ bool THING::CanRemove(MONSTER* monster)

bool THING::CanRemove(MONSTER* monster)
{
    if(monster == NULL || !monster->pack.worn(this))
	return TRUE;
    else if(IsCursed())
	return FALSE;

    return TRUE;
}

// @@@ long THING::worth(MONSTER* monster)

long THING::worth(MONSTER* monster)
{
    if(monster->knowledge(this) == 0)
	return ClassValue();
    else if(monster->know(this) == FALSE)
	return BaseValue();
    else if(IsCursed())
	return (BaseValue() * (long) sqr(MyValue)) / 10;
    else if(IsBlessed())
	return BaseValue() * (long) sqr(MyValue) * 10L;
    else
	return BaseValue() * (long) sqr(MyValue);
}

// @@@ void THING::effect(LEVEL*, MONSTER*)

void THING::effect(LEVEL*, MONSTER*)
{ }

// @@@ QUESTION* THING::DoEffect(LEVEL* level, PLAYER* player, CHANNEL*)

QUESTION* THING::DoEffect(LEVEL* level, PLAYER* player, CHANNEL*)
{
    effect(level, player);
    return NULL;
}

// @@@ void THING::TakeOff(MONSTER* monster)

void THING::TakeOff(MONSTER* monster)
{ monster->pack.TakeOff(this); }

// @@@ void THING::wear(MONSTER* monster)

void THING::wear(MONSTER* monster)
{ monster->pack.wear(this); }

// @@@ void THING::damage()

void THING::damage()
{
    if(!IsBlessed() && MyValue > 0)
	MyValue--;
}

// @@@ void THING::MoveAway(LEVEL* level, coor pos, int length)
//
// WARNING: object should not apear in the dungeon elsewhere

void THING::MoveAway(LEVEL* level, coor pos, int length)
{
    BUNCH* bunch = level->create(pos + coor(rnd(-length, length),
					    rnd(-length, length)));
    bunch->add(this);
}

// @@@ void THING::save(BINARY* binary)

void THING::save(BINARY* binary)
{
    OBJECT::save(binary);
    binary->PutChar(MyLetter);
    binary->PutChar(MyValue);
    binary->PutChar(holy);
    binary->PutChar(invisible);
    binary->PutLong(MyCount);
}

// @@@ THING::THING(BINARY* binary)

THING::THING(BINARY* binary) : (binary)
{
    MyLetter = binary->GetChar();
    MyValue = binary->GetChar();
    holy = binary->GetChar();
    invisible = binary->GetChar();
    MyCount = binary->GetLong();
}

// @@@ THING::THING(PARSE* parse)

THING::THING(PARSE* parse) : (parse)
{
    create();
    
    while(parse->good())
    {
	const char* string = parse->GetWord();

	if(strcmp(string, "LETTER") == EQUAL)
	{
	    parse->skip(INPUT_SPACE);
	    MyLetter = parse->GetKey();
	}
	else if(strcmp(string, "VALUE") == EQUAL)
	{
	    parse->skip(INPUT_SPACE);
	    MyValue = (char) parse->GetLong();
	}
	else if(strcmp(string, "CURSED") == EQUAL)
	    holy = CURSED;
	else if(strcmp(string, "BLESSED") == EQUAL)
	    holy = BLESSED;
	else if(strcmp(string, "INVISIBLE") == EQUAL)
	    invisible = TRUE;
	else if(strcmp(string, "COUNT") == EQUAL)
	{
	    parse->skip(INPUT_SPACE);
	    MyCount = parse->GetLong();
	}
	else if(strcmp(string, "BREAK") == EQUAL)
	    break;
	else
	    parse->wrong("THING::THING", format("Unknown field: %s", string));

	parse->skip(INPUT_COMMENT);
    }
    parse->skip(INPUT_COMMENT);
}

// @@@ THING::THING(THING* thing, long count)

THING::THING(THING* thing, long count)
{
    MyLetter = thing->MyLetter;
    MyValue = thing->MyValue;
    holy = thing->holy;
    invisible = thing->invisible;
    MyCount = count;
}    

// @@@ THING::THING(long count = 1L)

THING::THING(long count = 1L) : ()
{ create(count); }

// @@ MAGIC

// @@@ const char* MAGIC::name()

const char* MAGIC::name()
{
    STRING string = format("%s of %s", ClassName(), MagicName());
    const char* text = string.string();
    return text;
}

// @@@ const char* MAGIC::PluralName()

const char* MAGIC::PluralName()
{
    STRING string = format("%s of %s", PluralClassName(), MagicName());
    const char* text = string.string();
    return text;
}

// @@ TORCH

// @@@ void TORCH::create()

void TORCH::create()
{
    last = 0L;			// Not burning
    left = time();
    SetLetter('t');
}

// @@@ long TORCH::time()

long TORCH::time()
{
    return rnd(5 * 60 * (1 << (MyValue / 5)), 10 * 60 * (1 << MyValue / 5));
}

// @@@ comp TORCH::compare(OBJECT* object)

comp TORCH::compare(OBJECT* object)
{
    if(object->type() != WAND_TORCH)
	return THING::compare(object);
    else if(object->light() && !light())
	return GREATER_THAN;
    else if(light() && !object->light())
	return LESS_THAN;

    comp c = THING::compare(object);

    if(c == EQUAL)
	return ::compare(identity(), object->identity());
    else
	return c;
}

// @@@ void TORCH::apply(LEVEL* level, MONSTER* monster)

void TORCH::apply(LEVEL* level, MONSTER* monster)
{
    monster->UseMovePoint(25 - monster->experience());

    if(light())
    {
	monster->message(format("You extinguish %s.",
				::print(this, PRINT_OBJECT)));
	level->WatchMonster(monster, "%s extinguishes %s.", this);
	left -= (watch() - last);
	last = 0L;
	monster->pack.add(this);
    }
    else if(MyValue < 2)
    {
	monster->message("It burnes up immediately...");
	monster->pack.destroy(this);
    }
    else
    {
	monster->message(format("You set fire to %s.",
				::print(this, PRINT_OBJECT)));
	level->WatchMonster(monster, "%s sets fire to %s.", this);
	MyValue--;
	last = watch();
	left += time();
	monster->pack.add(this);
    }
}

// @@@ const char* TORCH::PrintActive(print_flag)

const char* TORCH::PrintActive(print_flag)
{
    if(light())
	return "burning";
    else
	return NULL;
}

// @@@ cont TORCH::event(LEVEL* level, coor pos, MONSTER* monster = NULL)

cont TORCH::event(LEVEL* level, coor pos, MONSTER* monster = NULL)
{
    if(light())
    {
	while(last + left < watch())
	{
	    if(MyValue < 1)
		break;
	    else
	    {
		left += time();
		if(IsBlessed())
		{
		    if(monster != NULL)
		    {
			monster->message
			    (format("Your %s glows momentarily up.",
				    ::print(this, PRINT_VALUE | PRINT_TYPE)));
			monster->identify(this);
		    }
		}
		else
		{
		    if(monster != NULL)
		    {
			monster->message(format("Your %s is growing faint.",
						::print(this,
							PRINT_VALUE |
							PRINT_TYPE)));
		    }
		    MyValue--;
		}
	    }
	}
    }
    return THING::event(level, pos, monster);
}

// @@@ void TORCH::save(BINARY* binary)

void TORCH::save(BINARY* binary)
{
    THING::save(binary);
    binary->PutLong(last);
    binary->PutLong(left);
}

// @@@ TORCH::TORCH(BINARY* binary)

TORCH::TORCH(BINARY* binary) : (binary)
{
    last = binary->GetLong();
    left = binary->GetLong();
}

// @@ WAND

// @@@ void WAND::cleanup(MONSTER* monster)

void WAND::cleanup(MONSTER* monster)
{
    MyValue = rnd(0, MyValue);

    if(MyValue < 1)
	monster->pack.destroy(this);
    else
	monster->pack.add(this);
}

// @@@ RAY* WAND::RayCursed()

RAY* WAND::RayCursed()
{
    bug.error("WAND::RayCursed", "virtual");
    return new RAY_DEATH(MyValue);
}

// @@@ RAY* WAND::RayNormal()

RAY* WAND::RayNormal()
{
    bug.error("WAND::RayNormal", "virtual");
    return new RAY_DEATH(MyValue);
}

// @@@ RAY* WAND::RayBlessed()

RAY* WAND::RayBlessed()
{
    bug.error("WAND::RayBlessed", "virtual");
    return new RAY_DEATH(MyValue);
}

// @@@ void WAND::apply(LEVEL* level, MONSTER* monster)

void WAND::apply(LEVEL* level, MONSTER* monster)
{
    monster->learn(this);

    // Tell the world
    monster->message(format("You zap %s.", ::print(this, PRINT_OBJECT)));
    level->WatchMonster(monster, "%s zaps %s.", this);

    switch(holy)
    {
    case CURSED:
	level->zap(monster, RayCursed(), monster->enemy(level)->position());
	break;
    case NORMAL:
	level->zap(monster, RayNormal(), monster->enemy(level)->position());
	break;
    case BLESSED:
	level->zap(monster, RayBlessed(), monster->enemy(level)->position());
	break;
    default:
	bug.error("WAND::apply", "default");
    }
    cleanup(monster);
}

// @@@ QUESTION* WAND::DoApply(LEVEL* level, PLAYER* player, CHANNEL* channel)

QUESTION* WAND::DoApply(LEVEL* level, PLAYER* player, CHANNEL* channel)
{
    player->learn(this);

    QUESTION* answer = NULL;

    // Tell the world
    player->message(format("You zap %s.",
			   ::print(this, PRINT_OBJECT, player)));
    level->WatchMonster(player, "%s zaps %s.", this);

    switch(holy)
    {
    case CURSED:
	answer = new QUESTION_ZAP(RayCursed(), channel);
	break;
    case NORMAL:
	answer = new QUESTION_ZAP(RayNormal(), channel);
	break;
    case BLESSED:
	answer = new QUESTION_ZAP(RayBlessed(), channel);
	break;
    default:
	bug.error("WAND::DoApply", "default");
    }
    cleanup(player);

    return answer;
}

// @@ POTION

// @@@ void POTION::apply(LEVEL* level, MONSTER* monster)

void POTION::apply(LEVEL* level, MONSTER* monster)
{
    monster->UseMovePoint(15);
    monster->UseFightPoint(5);
    
    monster->pack.destroy(this);
    monster->message(format("You quaff %s.", ::print(this, PRINT_OBJECT)));
    level->WatchMonster(monster, "%s quaffs %s.", this);
    effect(level, monster);
}

// @@ SCROLL

// @@@ void SCROLL::read(LEVEL* level, MONSTER* monster)

void SCROLL::read(LEVEL* level, MONSTER* monster)
{
    monster->UseMovePoint(75);
    monster->UseFightPoint(25);
    
    monster->pack.destroy(this);
    monster->message(format("You read %s.", ::print(this, PRINT_OBJECT)));
    level->WatchMonster(monster, "%s reads %s.", this);
    if(monster->IsConfused())
	monster->message("Being confused, you mispronounce the magic words.");
}

// @@@ void SCROLL::apply(LEVEL* level, MONSTER* monster)

void SCROLL::apply(LEVEL* level, MONSTER* monster)
{
    read(level, monster);
    effect(level, monster);
}

// @@@ QUESTION* SCROLL::DoApply(LEVEL* level,

QUESTION* SCROLL::DoApply(LEVEL* level, PLAYER* player, CHANNEL* channel)
{
    read(level, player);
    return DoEffect(level, player, channel);
}

// @@ RING

// @@@ void RING::apply(LEVEL* level, MONSTER* monster)

void RING::apply(LEVEL* level, MONSTER* monster)
{
    monster->UseMovePoint(5);
    monster->UseFightPoint(2);
    
    if(monster->pack.worn(this))
    {
	if(CanRemove(monster))
	{
	    monster->message(format("You remove %s.",
				    ::print(this, PRINT_OBJECT)));
	    level->WatchMonster(monster, "%s removes %s.", this);
	    TakeOff(monster);
	}
	else
	    monster->message(format("I can't remove %s.",
				    ::print(this, PRINT_THE)));
	
    }
    else if(monster->MayWearRing())
    {
	wear(monster);
	monster->message(format("You wear %s.", ::print(this, PRINT_OBJECT)));
	level->WatchMonster(monster, "%s wears %s.", this);
    }
    else
	monster->message("You have no free ring fingers left.");

    monster->pack.add(this);
}

// @@ AMULET

// @@@ void AMULET::apply(LEVEL* level, MONSTER* monster)

void AMULET::apply(LEVEL* level, MONSTER* monster)
{
    monster->UseMovePoint(10);
    monster->UseFightPoint(5);
    
    if(monster->pack.worn(this))
    {
	if(CanRemove(monster))
	{
	    monster->message(format("You remove %s.",
				    ::print(this, PRINT_OBJECT)));
	    level->WatchMonster(monster, "%s removes %s.", this);
	    TakeOff(monster);
	}
	else
	    monster->message(format("I can't remove %s.",
				    ::print(this, PRINT_THE)));
	
    }
    else
    {
	wear(monster);
	monster->message(format("You wear %s.", ::print(this, PRINT_OBJECT)));
	level->WatchMonster(monster, "%s wears %s.", this);
    }
    monster->pack.add(this);
}

// @@ WEAPON

// @@@ void WEAPON::apply(LEVEL* level, MONSTER* monster)

void WEAPON::apply(LEVEL* level, MONSTER* monster)
{
    monster->UseMovePoint(15);
    monster->UseFightPoint(5);
    
    if(monster->pack.worn(this))
    {
	if(CanRemove(monster))
	{
	    monster->message(format("You unwield %s.",
				    ::print(this, PRINT_OBJECT)));
	    level->WatchMonster(monster, "%s unwields %s.", this);
	    TakeOff(monster);
	}
	else
	    monster->message(format("%s is welded to your hand!",
				    ::print(this, PRINT_THE | PRINT_UPPER)));
    }
    else
    {
	OBJECT* object = monster->pack.weapon();

	if(object == this)
	    monster->message(format("You already wield %s!",
				    ::print(this, PRINT_TYPE | PRINT_THE)));
	else if(object != NULL && !object->CanRemove(monster))
	    monster->message(format("You must find a way to unwield %s first.",
				    ::print(object, PRINT_TYPE | PRINT_THE)));
	else
	{
	    if(object != NULL)
		object->TakeOff(monster);

	    wear(monster);
	    monster->message(format("You wield %s.",
				    ::print(this, PRINT_OBJECT)));
	    level->WatchMonster(monster, "%s wields %s.", this);
	    if(IsCursed())
		monster->message("It welds itself to your hand!");
	}
    }
    monster->pack.add(this);
}

// @@ ARMOR

// @@@ void ARMOR::apply(LEVEL* level, MONSTER* monster)

void ARMOR::apply(LEVEL* level, MONSTER* monster)
{
    monster->UseMovePoint(75);
    monster->UseFightPoint(5);
    
    if(monster->pack.worn(this))
    {
	if(CanRemove(monster))
	{
	    monster->message(format("You take of %s.",
				    ::print(this, PRINT_OBJECT)));
	    level->WatchMonster(monster, "%s takes of %s.", this);
	    TakeOff(monster);
	}
	else
	    monster->message(format("%s is cursed!",
				  ::print(this, PRINT_THE | PRINT_UPPER)));
    }
    else
    {
	OBJECT* object = monster->pack.armor();

	if(object == this)
	    monster->message(format("You already wear %s!",
				    ::print(this, PRINT_TYPE | PRINT_THE)));
	else if(object != NULL && !object->CanRemove(monster))
	    monster->message(format("You must take of %s first.",
				    ::print(object, PRINT_TYPE | PRINT_THE)));
	else
	{
	    if(object != NULL)
		object->TakeOff(monster);
	    
	    wear(monster);
	    monster->message(format("You wear %s.",
				    ::print(this, PRINT_OBJECT)));
	    level->WatchMonster(monster, "%s wears %s.", this);
	}
    }
    monster->pack.add(this);
}

// @@@ cont ARMOR::absorb(int* power, int* type,

cont ARMOR::absorb(int* power, damage_type* type,
		   OBJECT* object, MONSTER* attack, MONSTER* defence)
{
    // Reduce the power
    int factor = protect(*power, type) * MyValue;

    if(rnd.chance(10))
	// Sometime the armor is unimportant!
	*power = rnd(*power);
    else if(rnd(*power) < rnd(MyLog(factor) + defence->experience()))
	// Just ignore most weak blows
	*power = 0;
    else if(factor < 1000 * MaxValue)
    {
	// Take some per thousand of the damage
	long l = *power;
	l *= 1000L * MaxValue - factor;
	l /= 1000L * MaxValue;
	*power = (int) l;
    }
    else
	// Remove all the damage
	*power = 0;
    
    // Damage weapon and armor
    if(rnd(object->strength() * 250) < rnd(strength() * *power))
    {
	object->damage();
	attack->pack.touch();
    }
    if(rnd(*power) > rnd(MyValue * strength() * 10))
    {
	damage();
	defence->pack.touch();
    }
    return CONTINUE;
}

// @@ FOOD

// @@@ void FOOD::eat(LEVEL* level, MONSTER* monster)

void FOOD::eat(LEVEL* level, MONSTER* monster)
{
    monster->UseMovePoint(FoodValue() / 60);
    monster->UseFightPoint(1);
    
    monster->pack.destroy(this);
    monster->message(format("You eat %s.", ::print(this, PRINT_OBJECT)));
    level->WatchMonster(monster, "%s eats %s.", this);

    if(monster->feed(FoodValue(), IsBlessed() ? FORCE : CHECK) == DONE)
    {
	monster->message(format("You choke over %s.",
				::print(this, PRINT_OBJECT)));
	level->WatchMonster(monster, "%s chokes over %s.", this);
	monster->kill(format("choked over %s",
			     ::print(this, PRINT_OBJECT)), level);
    }
}

// @@@ const char* FOOD::PrintHoly()

const char* FOOD::PrintHoly()
{
    if(IsBlessed())
	return "fresh";
    else if(IsCursed())
	return "poisonous";
    else
	return NULL;
}

// @@@ void FOOD::apply(LEVEL* level, MONSTER* monster)

void FOOD::apply(LEVEL* level, MONSTER* monster)
{
    eat(level, monster);

    if(!monster->dead())
	effect(level, monster);
}

// @@@ QUESTION* FOOD::DoApply(LEVEL* level, PLAYER* player, CHANNEL* channel)

QUESTION* FOOD::DoApply(LEVEL* level, PLAYER* player, CHANNEL* channel)
{
    eat(level, player);

    if(!player->dead())
	return DoEffect(level, player, channel);

    return NULL;
}

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
